package org.xbib.interlibrary.catalog.matching.chronology;

import java.time.LocalDate;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Parsing patterns of enumeration and chronology.
 * Rules are given by Zeitschriftendatenbank
 * <a href="http://support.d-nb.de/iltis/katricht/zdb/8032.pdf">...</a>
 */
public class EnumerationAndChronologyHelper {

    private static final Integer currentYear = LocalDate.now().getYear();

    // Jahrgänge, mit Zählung, offen
    //
    // 1 = Zählung
    // 2 = Berichtsjahr
    // 3 = Erscheinungsjahr (yyyy), (yyyy/yy), (yyyy/yyyy)
    private static final Pattern[] open_vyy = {
            // [2.]1979(1981) -
            // 115.1921/22(1923) -
            Pattern.compile("\\[(.*?)\\.\\]\\s*(\\d{4})\\((\\d{0,4})\\).*\\-\\s*$"),
            Pattern.compile("(.*?)\\.\\s*(\\d{4}/?\\d{0,4})\\((\\d{4}/?\\d{0,4})\\)\\s*\\-\\s*$")
    };
    // Jahrgänge, mit Zählung, mit Berichtsjahr, offen
    // i.yyyy -
    // 1 = Zählung
    // 2 = Erscheinungsjahr
    private static final Pattern[] open_vy = {
            // 1.1971 -
            // 2.1947,15.Mai -
            Pattern.compile("(\\d+)\\.\\s*(\\d{4}/?\\d{0,4}).*?\\-\\s*$"),
            // 63.2011,Okt. -
            Pattern.compile("(\\d{0,4})\\.\\s*(\\d{4}).*?\\-\\s*$"),
            Pattern.compile("(.*?)\\.\\s*(\\d{4}/?\\d{0,4}).*\\-\\s*$")
    };
    // yyyy,v -
    private static final Pattern[] open_yv = {
            Pattern.compile("(\\d{4}/?\\d{0,4}),(.*?)\\s*\\-\\s*$")
    };
    // yyyy[/nnnn](yyyy[/nnnn]) -
    private static final Pattern[] open_yy = {
            Pattern.compile("(\\d{4}/?\\d{0,4})\\((\\d{4})\\)\\s*\\-\\s*$")
    };
    // yyyy -
    private static final Pattern[] open_y = {
            // 1971 -
            Pattern.compile("(\\d{4}/?\\d{0,4}).*\\-\\s*$"),
            // 1963,21(22.Mai) -
            Pattern.compile("(\\d{4}).*\\-\\s*$"),
            // [19]51,1 - [19]52,5
            Pattern.compile("(\\[\\d{2,4}\\]\\d{2,4}).*\\-\\s*$"),
            // Jahrgang 01, 01 (2022)-
            Pattern.compile(".*\\((\\d{4})\\).*\\-\\s*$")
    };
    // Jahrgänge, mit zwei Zählungen, jeweils Berichtsjahr/Erscheinungsjahr, geschlossen (6 Gruppen)
    // 115.1921/22(1923) - 116.1922/23(1924)
    // 1 = Zählung
    // 2 = Berichtsjahr
    // 3 = Erscheinungsjahr (yyyy), (yyyy/yy), (yyyy/yyyy)
    // 4 = Zählung
    // 5 = Berichtsjahr
    // 6 = Erscheinungsjahr (yyyy), (yyyy/yy), (yyyy/yyyy)
    private static final Pattern[] closed_range_vyy_vyy = {
            // [1.]1985(1984) - [29.]2013(2012)
            Pattern.compile("\\[(.*?)\\.]\\s*(\\d{4}/?\\d{0,4})\\((\\d{4}/?\\d{0,4})\\)\\s*\\-"
                    + "\\s*\\[(.*?)\\.]\\s*(\\d{4}/?\\d{0,4})\\((\\d{4}/?\\d{0,4})\\)"),
            Pattern.compile("(.*?)\\.\\s*(\\d{4}/?\\d{0,4})\\((\\d{4}/?\\d{0,4})\\)\\s*\\-"
                    + "\\s*(.*?)\\.\\s*(\\d{4}/?\\d{0,4})\\((\\d{4}/?\\d{0,4})\\)")
    };
    // Jahrgänge, mit zwei Zählungen, vorne Berichtsjahr/Erscheinungsjahr, geschlossen (5 Gruppen)
    // 115.1921/22(1923) - 1125.1937
    // 1 = Zählung
    // 2 = Berichtsjahr
    // 3 = Erscheinungsjahr (yyyy), (yyyy/yy), (yyyy/yyyy)
    // 4 = Zählung
    // 5 = Berichtsjahr
    private static final Pattern[] closed_range_vyy_vy = {
            Pattern.compile("(.*?)\\.\\s*(\\d{4}/?\\d{0,4})\\((\\d{4}/?\\d{0,4})\\)\\s*\\-\\s*(.*?)\\.\\s*(\\d{4}/?\\d{0,4})")
    };
    // Jahrgänge, mit zwei Zählungen, hinten Berichtsjahr/Erscheinungsjahr, geschlossen (5 Gruppen)
    // 10.1991 - 55.1993/94(1994)
    // 39(=48).1955 - 59(=68).1975(1976)
    // 1 = Zählung
    // 2 = Berichtsjahr
    // 3 = Zählung
    // 4 = Berichtsjahr
    // 5 = Erscheinungsjahr (yyyy), (yyyy/yy), (yyyy/yyyy)
    private static final Pattern[] closed_range_vy_vyy = {
            Pattern.compile("(.*?)\\.\\s*(\\d{4}/?\\d{0,4})\\s*\\-\\s*(.*?)\\.\\s*(\\d{4}/?\\d{0,4})\\((\\d{0,4})\\)")
    };
    // i.yyyy - i.yyyy (4 Gruppen)
    private static final Pattern[] closed_range_vy_vy = {
            // 6.1961/64 - 31.1970
            // 1.1963 - 12.1972
            // 115.1921/22(1923) - 1125.1937
            // 3.1858,6 - 24.1881,3
            // 1.1960 - 5.1963; 11.1964; 23.1971 -
            Pattern.compile("(.+?)\\.\\s*(\\d{4}/?\\d{0,4})\\s*\\-\\s*(.+?)\\.\\s*(\\d{4}/?\\d{0,4})"),
            Pattern.compile("(.+?)\\.\\s*(\\d{4}/?\\d{0,4}).*\\-\\s*(.+?)\\.\\s*(\\d{4}/?\\d{0,4})"),
            Pattern.compile("(.+?)\\.\\s*(\\d{4}/?\\d{0,4}),.*?\\s*\\-\\s*(.+?)\\.\\s*(\\d{4}/?\\d{0,4})"),
            Pattern.compile("(.+?)\\.\\s*(\\d{4}/?\\d{0,4}\\(\\d{0,4}\\))\\s*\\-\\s*(.+?)\\.\\s*(\\d{4}/?\\d{0,4}\\(\\d{0,4}\\))"),
            Pattern.compile("(.+?)\\.\\s*(\\d{4}/?\\d{0,4}\\(\\d{0,4}\\)),.*?\\s*\\-\\s*(.+?)\\.\\s*(\\d{4}/?\\d{0,4}\\(\\d{0,4}\\))"),
            // [19]81/82 - [19]83
            Pattern.compile("(.+?)\\.\\s*(\\[\\d{2,4}\\]\\d{2,4})(/\\d{2,4}).*\\-\\s*(.+?)\\.\\s*(\\[\\d{2,4}\\]\\d{2,4})(/\\d{2,4})"),
            // 1.[19]51,1 - 1.[19]52,5
            Pattern.compile("(.+?)\\.\\s*(\\[\\d{2,4}\\]\\d{2,4}).*\\-\\s*(.+?)\\.\\s*(\\[\\d{2,4}\\]\\d{2,4})"),
            // [1]. 1987 - 25.2011 = Nr. 1 - 74/75
            Pattern.compile("(\\[\\d+\\])\\.\\s*(\\d{4}).*\\-\\s*(\\d+)\\.\\s*(\\d{4})")
    };
    private static final Pattern[] closed_range_y_vy = {
            // 1981 - 54/55.2005"
            Pattern.compile("(\\d{4}/?\\d{0,4})\\s*\\-\\s*(\\d+/\\d+)\\.\\s*(\\d{4}/?\\d{0,4})"),
            // 1981 - 55.2005"
            Pattern.compile("(\\d{4}/?\\d{0,4})\\s*\\-\\s*(\\d+)\\.\\s*(\\d{4}/?\\d{0,4})"),
            // 1981,31 - 25.1997
            Pattern.compile("(\\d{4}/?\\d{0,4}),.*?\\s*\\-\\s*(\\d+)\\.\\s*(\\d{4}/?\\d{0,4})")
    };
    // yyyy - yyyy
    private static final Pattern[] closed_range_y_y = {
            // [19]63 - [19]72
            Pattern.compile("\\[(\\d{2})\\](\\d{2}/?\\d{0,4})\\s*\\-\\s*\\[(\\d{2})\\](\\d{2}/?\\d{0,4})"),
            // 1963 - 1972
            Pattern.compile("(\\d{4}/?\\d{0,4}).*\\-\\s*(\\d{4}/?\\d{0,4})"),
    };
    // yyyy,v - yyyy,v
    private static final Pattern[] closed_range_yv_yv = {
            // 1951,1 - 1952,5
            Pattern.compile("(\\d{4}),(.*?)\\s*\\-\\s*(\\d{4}),(.*?)"),
            // [19]51,1 - [19]52,5
            Pattern.compile("(\\[\\d{2,4}\\]\\d{2,4}/?\\d{0,4}),(.*?)\\-\\s*(\\[\\d{2,4}\\]\\d{2,4}/?\\d{2,4}),(.*?)")

    };
    // i.yyyy - yyyy,v
    private static final Pattern[] closed_range_vy_yv = {
            Pattern.compile("(.*?)\\.\\s*(\\d{4})\\s*\\-\\s*(\\d{4}),(.*?)")
    };
    // Jahrgänge, ohne Zählungen, jeweils Berichtsjahr(doppel)/Erscheinungsjahr, geschlossen
    // yyyy/...(yyyy) - yyyy/...(yyyy)
    // 1 = Berichtsjahr
    // 2 = Erscheinungsjahr
    // 3 = Berichtsjahr
    // 4 = Erscheinungsjahr
    private static final Pattern[] closed_range_yy_yy = {
            // 2004/05(2004) - 2006/07(2006)
            Pattern.compile("(\\d{4})/.*?\\((\\d{4})\\)\\s*\\-\\s*(\\d{4})/.*?\\((\\d{4})\\)")
    };

    // ein Jahrgang, mit Zählung, ohne Berichtsjahr, geschlossen
    // 1 = Zählung
    // 2 = Erscheinungsjahr
    // 3 = Berichtsjahr (yyyy), (yyyy/yy), (yyyy/yyyy)
    private static final Pattern[] closed_point_vyy = {
            // 2.1938/40(1942)
            // 9.1996/97(1997)
            Pattern.compile("(\\d+)\\.\\s*(\\d{4}/?\\d{0,4})\\((\\d{4}/?\\d{0,4})\\)")
    };
    // ein Jahrgang, mit Zählung, ohne Berichtsjahr, geschlossen
    // i.yyyy
    // 1 = Zählung
    // 2 = Erscheinungsjahr
    private static final Pattern[] closed_point_vy = {
            // 1.1970/71
            Pattern.compile("(\\d+)\\.\\s*(\\d{4})(/\\d{0,4})"),
            // 1.1970
            // 2.1970,3
            // 4.1961,Aug.
            // 3.1971,Jan./Febr.
            // 1.1970; 3.1972; 7.1973
            Pattern.compile("(\\d+)\\.\\s*(\\d{4})"),
            // 1.[19]93,1
            Pattern.compile("(\\d+)\\.\\s*(\\[?\\d{2,4}\\]?\\d{2,4}/?\\d{0,4})"),
            // 1.1981=1401
            Pattern.compile("(\\d+)\\.\\s*(\\d{4}/?\\d{0,4})\\s*=")
    };
    // ein Jahrgang, ohne Zählung, geschlossen
    // 1 = Berichtsjahr
    // 2 = Erscheinungsjahr
    private static final Pattern[] closed_point_yy = {
            // 1961/62(1963)
            // 1965/70(1971/72)
            Pattern.compile("(\\d{4}/?\\d{0,4})\\((\\d{4}/?\\d{0,4})\\)"),
            // 1965/70(1971/72)
            Pattern.compile("(\\d{4})(/?\\d{0,4})\\((\\d{4})(/?\\d{0,4})\\)"),
            // 1.5678=[1917/18]
            Pattern.compile("\\d+\\.\\s*(\\d{4}/?\\d{0,4})\\s*=\\s*\\[(\\d{4}/?\\d{2,4})\\]"),
            // 1.5678=1917/18
            Pattern.compile("\\d+\\.\\s*(\\d{4}/?\\d{0,4})\\s*=\\s*(\\d{4}/?\\d{2,4})"),
            // 5717=[1956/57]
            Pattern.compile("(\\d{4}/?\\d{0,4})\\s*=\\s*\\[(\\d{4}/?\\d{0,4})\\]"),
            // 5717=1956
            Pattern.compile("(\\d{4}/?\\d{0,4})\\s*=\\s*(\\d{4}/?\\d{0,4})"),
    };
    // ein Jahrgang, ohne Zählung, geschlossen
    // 1 = Erscheinungsjahr
    private static final Pattern[] closed_point_y = {
            // An V= [1796/97]
            // ...=[1917/18]
            Pattern.compile("=\\s*\\[(\\d{4}/?\\d{0,4})\\]"),
            // ...=1956/57
            Pattern.compile("=\\s*(\\d{4}/?\\d{0,4})"),
            // 1921/22=...
            Pattern.compile("(\\d{4})(/?\\d{0,4})\\s*="),
            // 1965/66
            // 1968/70
            // 1961/62(1963)
            // 1961/62(1962)
            // WS 1948/49
            Pattern.compile("(\\d{4})(/?\\d{0,4})"),
            // 1921=1339
            // 1961
            // 1992,14140(12. März)
            // SS 1922
            Pattern.compile("(\\d{4})")
    };

    private final List<Integer> begin;

    private final List<Integer> end;

    private final List<String> beginVolume;

    private final List<String> endVolume;

    private final List<Integer> beginPubDate;

    private final List<Integer> endPubDate;

    private final List<Boolean> open;

    public EnumerationAndChronologyHelper() {
        this.begin = new LinkedList<>();
        this.end = new LinkedList<>();
        this.beginVolume = new LinkedList<>();
        this.endVolume = new LinkedList<>();
        this.beginPubDate = new LinkedList<>();
        this.endPubDate = new LinkedList<>();
        this.open = new LinkedList<>();
    }

    public void parse(String content) {
        parseGroup(content);
    }

    public List<Integer> getBegin() {
        return begin;
    }

    public List<Integer> getEnd() {
        return end;
    }

    public List<String> getBeginVolume() {
        return beginVolume;
    }

    public List<String> getEndVolume() {
        return endVolume;
    }

    public List<Integer> getBeginPubDate() {
        return beginPubDate;
    }

    public List<Integer> getEndPubDate() {
        return endPubDate;
    }

    public List<Boolean> getOpen() {
        return open;
    }

    public Collection<Integer> getDates() {
        return getDates(null);
    }

    public Collection<Integer> getDates(String content) {
        Set<Integer> set = new TreeSet<>();
        int max = content != null ? parseGroup(content) : begin.size();
        for (int i = 0; i < max; i++) {
            Integer b = begin.get(i);
            Integer e = end.get(i);
            Integer beginyear = beginPubDate.get(i);
            Integer endyear = endPubDate.get(i);
            Boolean isopen = open.get(i);
            if (beginyear != null) {
                if (endyear != null) {
                    for (int y = beginyear; y <= endyear; y++) {
                        set.add(y);
                    }
                } else if (isopen != null && isopen) {
                    for (int y = beginyear; y <= currentYear; y++) {
                        set.add(y);
                    }
                } else {
                    set.add(beginyear);
                }
            } else if (endyear != null) {
                set.add(endyear);
            }
            if (b != null) {
                if (e != null) {
                    for (int y = b; y <= e; y++) {
                        set.add(y);
                    }
                } else if (isopen != null && isopen) {
                    for (int y = b; y <= currentYear; y++) {
                        set.add(y);
                    }
                } else {
                    set.add(b);
                }
            } else if (e != null) {
                set.add(e);
            }
        }
        return set;
    }

    public boolean hasYear(Integer year, String values) {
        if (year == null) {
            return true;
        }
        int max = parseGroup(values);
        for (int i = 0; i < max; i++) {
            Integer b = begin.get(i);
            Integer e = end.get(i);
            Integer beginyear = beginPubDate.get(i);
            Integer endyear = endPubDate.get(i);
            Boolean isopen = open.get(i);
            if (beginyear != null && endyear != null && beginyear <= year && year <= endyear) {
                return true;
            }
            if (beginyear != null && beginyear <= year && isopen != null && isopen) {
                return true;
            }
            if (b != null && e != null && b <= year && year <= e) {
                return true;
            }
            if (b != null && b <= year && isopen != null && isopen) {
                return true;
            }
        }
        return false;
    }

    public void parseMovingWall(List<Pattern> movingWallPatterns, String content) {
        int i = begin.size();
        if (movingWallPatterns != null) {
            for (Pattern p : movingWallPatterns) {
                Matcher m = p.matcher(content);
                if (m.find()) {
                    Integer b = currentYear - Integer.parseInt(m.group(1));
                    begin.add(i, b);
                    end.add(i, null);
                    beginVolume.add(i, null);
                    endVolume.add(i, null);
                    beginPubDate.add(i, b);
                    endPubDate.add(i, null);
                    open.add(i, true);
                    i++;
                }
            }
        }
    }

    private int parseGroup(String values) {
        if (values == null) {
            return 0;
        }
        int i = 0;
        for (String value : values.split(";")) {
            i = parseGroup(i, value);
        }
        return i;
    }

    private int parseGroup(int i, String value) {
        boolean found = false;
        for (Pattern p : open_vyy) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                begin.add(i, sanitizeDate(m.group(2)));
                end.add(i, null);
                beginVolume.add(i, trim(m.group(1)));
                endVolume.add(i, null);
                beginPubDate.add(i, sanitizeDate(m.group(3)));
                endPubDate.add(i, null);
                open.add(i, true);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : open_yv) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                Integer b = sanitizeDate(m.group(1));
                String bvol = trim(m.group(2));
                begin.add(i, b);
                beginPubDate.add(i, b);
                beginVolume.add(i, bvol);
                end.add(i, null);
                endPubDate.add(i, null);
                endVolume.add(i, null);
                open.add(i, true);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : open_vy) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                Integer b = sanitizeDate(m.group(2));
                String bvol = trim(m.group(1));
                begin.add(i, b);
                beginPubDate.add(i, b);
                beginVolume.add(i,  bvol);
                end.add(i, null);
                endPubDate.add(i, null);
                endVolume.add(i, null);
                open.add(i, true);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : open_yy) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                Integer b1 = sanitizeDate(m.group(1));
                Integer b2 = sanitizeDate(m.group(2));
                begin.add(i, b1);
                beginPubDate.add(i, b2);
                beginVolume.add(i, null);
                end.add(i, null);
                endPubDate.add(i, null);
                endVolume.add(i, null);
                open.add(i, true);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : open_y) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                Integer b = sanitizeDate(m.group(1));
                begin.add(i, b);
                end.add(i, null);
                beginVolume.add(i, null);
                endVolume.add(i, null);
                beginPubDate.add(i, b);
                endPubDate.add(i, null);
                open.add(i, true);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_range_vyy_vyy) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                begin.add(i, sanitizeDate(m.group(2)));
                beginPubDate.add(i, sanitizeDate(m.group(3)));
                beginVolume.add(i, trim(m.group(1)));
                end.add(i, sanitizeDate(m.group(5)));
                endPubDate.add(i, sanitizeDate(m.group(6)));
                endVolume.add(i, trim(m.group(4)));
                open.add(i, false);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_range_vyy_vy) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                begin.add(i, sanitizeDate(m.group(2)));
                end.add(i, sanitizeDate(m.group(5)));
                beginVolume.add(i, trim(m.group(1)));
                endVolume.add(i, trim(m.group(4)));
                beginPubDate.add(i, sanitizeDate(m.group(3)));
                endPubDate.add(i, sanitizeDate(m.group(5)));
                open.add(i, false);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_range_vy_vyy) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                begin.add(i, sanitizeDate(m.group(2)));
                end.add(i, sanitizeDate(m.group(4)));
                beginVolume.add(i, trim(m.group(1)));
                endVolume.add(i, trim(m.group(3)));
                beginPubDate.add(i, sanitizeDate(m.group(2)));
                endPubDate.add(i, sanitizeDate(m.group(5)));
                open.add(i, false);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_range_vy_vy) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                List<Integer> l1 = fractionDate(m.group(2));
                List<Integer> l2 = fractionDate(m.group(4));
                int bIndex = l1.getFirst();
                Integer b = sanitizeDate(bIndex);
                begin.add(i, b);
                beginVolume.add(i, trim(m.group(1)));
                beginPubDate.add(i, b);
                int eIndex = !l2.isEmpty() ? l2.getLast() : -1;
                if (eIndex >= 0) {
                    Integer e = sanitizeDate(eIndex);
                    end.add(i, e);
                    endPubDate.add(i, e);
                    endVolume.add(i, trim(m.group(3)));
                } else {
                    end.add(i, b);
                    endPubDate.add(i, b);
                    endVolume.add(i, trim(m.group(3)));
                }
                open.add(i, false);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_range_y_vy) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                List<Integer> l1 = fractionDate(m.group(1));
                List<Integer> l2 = fractionDate(m.group(3));
                Integer d1 = sanitizeDate(l1.getFirst());
                Integer d2 = sanitizeDate(l2.getLast());
                begin.add(i, d1);
                end.add(i, d2);
                beginVolume.add(i, null);
                endVolume.add(i, trim(m.group(2)));
                beginPubDate.add(i, d1);
                endPubDate.add(i, d2);
                open.add(i, false);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_range_yy_yy) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                begin.add(i, sanitizeDate(m.group(1)));
                end.add(i, sanitizeDate(m.group(3)));
                beginVolume.add(i, null);
                endVolume.add(i, null);
                beginPubDate.add(i, sanitizeDate(m.group(2)));
                endPubDate.add(i, sanitizeDate(m.group(4)));
                open.add(i, false);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_range_vy_yv) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                Integer b = sanitizeDate(m.group(2));
                Integer e = sanitizeDate(m.group(3));
                begin.add(i, b);
                end.add(i, e);
                beginVolume.add(i, trim(m.group(1)));
                endVolume.add(i, trim(m.group(4)));
                beginPubDate.add(i, b);
                endPubDate.add(i, e);
                open.add(i, false);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_range_yv_yv) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                Integer b = sanitizeDate(m.group(1));
                Integer e = sanitizeDate(m.group(3));
                begin.add(i, b);
                end.add(i, e);
                beginVolume.add(i, trim(m.group(2)));
                endVolume.add(i, trim(m.group(4)));
                beginPubDate.add(i, b);
                endPubDate.add(i, e);
                open.add(i, false);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_range_y_y) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                if (m.groupCount() == 4) {
                    List<Integer> l1 = fractionDate(m.group(1) + m.group(2));
                    List<Integer> l2 = fractionDate(m.group(3) + m.group(4));
                    Integer b = sanitizeDate(l1.getFirst());
                    Integer e = sanitizeDate(l2.getLast());
                    begin.add(i, b);
                    end.add(i, e);
                    beginVolume.add(i, null);
                    endVolume.add(i, null);
                    beginPubDate.add(i, b);
                    endPubDate.add(i, e);
                    open.add(i, false);
                } else {
                    List<Integer> l1 = fractionDate(m.group(1));
                    List<Integer> l2 = fractionDate(m.group(2));
                    Integer b = sanitizeDate(l1.getFirst());
                    Integer e = sanitizeDate(l2.getLast());
                    begin.add(i, b);
                    end.add(i, e);
                    beginVolume.add(i, null);
                    endVolume.add(i, null);
                    beginPubDate.add(i, b);
                    endPubDate.add(i, e);
                    open.add(i, false);
                }
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_point_vyy) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                begin.add(i, sanitizeDate(m.group(2)));
                end.add(i, null);
                beginVolume.add(i, trim(m.group(1)));
                endVolume.add(i, null);
                beginPubDate.add(i, sanitizeDate(m.group(3)));
                endPubDate.add(i, null);
                open.add(i, false);
                i++;
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_point_yy) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                Integer b = sanitizeDate(m.group(1));
                List<Integer> l = fractionDate(m.group(2));
                begin.add(i, b);
                beginPubDate.add(i, sanitizeDate(l.getFirst()));
                beginVolume.add(i, null);
                end.add(i, null);
                endPubDate.add(i, null);
                endVolume.add(i, null);
                open.add(i, false);
                i++;
                if (l.size() > 1) {
                    b = sanitizeDate(l.get(1));
                    begin.add(i, b);
                    beginPubDate.add(i, b);
                    beginVolume.add(i,  null);
                    end.add(i, null);
                    endPubDate.add(i, null);
                    endVolume.add(i, null);
                    open.add(i, false);
                    i++;
                }
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_point_vy) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                if (m.groupCount() == 3) {
                    List<Integer> l = fractionDate(m.group(2) + m.group(3));
                    Integer b = sanitizeDate(l.getFirst());
                    String bvol =  trim(m.group(1));
                    begin.add(i, b);
                    end.add(i, null);
                    beginVolume.add(i, bvol);
                    endVolume.add(i, null);
                    beginPubDate.add(i, b);
                    endPubDate.add(i, null);
                    open.add(i, false);
                    i++;
                    if (l.size() > 1) {
                        b = sanitizeDate(l.get(1));
                        begin.add(i, b);
                        end.add(i, null);
                        beginVolume.add(i, bvol);
                        endVolume.add(i, null);
                        beginPubDate.add(i, b);
                        endPubDate.add(i, null);
                        open.add(i, false);
                        i++;
                    }
                } else {
                    String bvol = trim(m.group(1));
                    Integer b = sanitizeDate(m.group(2));
                    begin.add(i, b);
                    beginPubDate.add(i, b);
                    beginVolume.add(i,  bvol);
                    end.add(i, null);
                    endPubDate.add(i, null);
                    endVolume.add(i, null);
                    open.add(i, false);
                    i++;
                }
                break;
            }
        }
        if (found) {
            return i;
        }
        for (Pattern p : closed_point_y) {
            Matcher m = p.matcher(value);
            found = m.find();
            if (found) {
                if (m.groupCount() == 4) {
                    List<Integer> l1 = fractionDate(m.group(1) + m.group(2));
                    List<Integer> l2 = fractionDate(m.group(3) + m.group(4));
                    Integer b = sanitizeDate(l1.getFirst());
                    begin.add(i, b);
                    end.add(i, null);
                    beginVolume.add(i, null);
                    endVolume.add(i, null);
                    beginPubDate.add(i, b);
                    endPubDate.add(i, null);
                    open.add(i, false);
                    i++;
                    if (l1.size() > 1) {
                        b = sanitizeDate(l1.get(1));
                        begin.add(i, b);
                        end.add(i, null);
                        beginVolume.add(i, null);
                        endVolume.add(i, null);
                        beginPubDate.add(i, b);
                        endPubDate.add(i, null);
                        open.add(i, false);
                        i++;
                    }
                    b = sanitizeDate(l2.getFirst());
                    begin.add(i, b);
                    end.add(i, null);
                    beginVolume.add(i, null);
                    endVolume.add(i, null);
                    beginPubDate.add(i, b);
                    endPubDate.add(i, null);
                    open.add(i, false);
                    i++;
                    if (l2.size() > 1) {
                        b = sanitizeDate(l2.get(1));
                        begin.add(i, b);
                        end.add(i, null);
                        beginVolume.add(i, null);
                        endVolume.add(i, null);
                        beginPubDate.add(i, b);
                        endPubDate.add(i, null);
                        open.add(i, false);
                        i++;
                    }
                } else if (m.groupCount() == 2) {
                    List<Integer> l = fractionDate(m.group(1) + m.group(2));
                    Integer b = sanitizeDate(l.getFirst());
                    begin.add(i, b);
                    end.add(i, null);
                    beginVolume.add(i, null);
                    endVolume.add(i, null);
                    beginPubDate.add(i, b);
                    endPubDate.add(i, null);
                    open.add(i, false);
                    i++;
                    if (l.size() > 1) {
                        b = sanitizeDate(l.get(1));
                        begin.add(i, b);
                        end.add(i, null);
                        beginVolume.add(i, null);
                        endVolume.add(i, null);
                        beginPubDate.add(i, b);
                        endPubDate.add(i, null);
                        open.add(i, false);
                        i++;
                    }
                } else {
                    Integer b = sanitizeDate(m.group(1));
                    begin.add(i, b);
                    end.add(i, null);
                    beginVolume.add(i, null);
                    endVolume.add(i, null);
                    beginPubDate.add(i, b);
                    endPubDate.add(i, null);
                    open.add(i, false);
                    i++;
                }
                break;
            }
        }
        return i;
    }

    private Integer sanitizeDate(String date) {
        int pos = date.indexOf('/');
        if (pos > 0) {
            if (pos > 4) {
                return null; // invalid date
            }
            String s = date.substring(0, pos).replaceAll("[^\\d]", "");
            return sanitizeDate(Integer.parseInt(s));
        } else {
            // remove non-numeric characters
            String s = date.replaceAll("[^\\d]", "");
            return s.length() == 4 ? sanitizeDate(Integer.parseInt(s)) : null;
        }
    }

    private Integer sanitizeDate(Integer date) {
        return date != null && date > 1500 && date <= currentYear ? date : null;
    }

    private String trim(String string) {
        return string.trim();
    }

    private List<Integer> fractionDate(String date) {
        List<Integer> dates = new LinkedList<>();
        int pos = date.indexOf('/');
        if (pos > 0) {
            String s = date.substring(0, pos).replaceAll("[^\\d]", "");
            int base = Integer.parseInt(s);
            dates.add(base);
            int frac = 0;
            try {
                frac = Integer.parseInt(date.substring(pos + 1));
            } catch (NumberFormatException e) {
                // not important
            }
            if (frac >= 100 && frac <= currentYear) {
                dates.add(frac);
            } else if (frac > 0 && frac < 100) {
                // two digits frac, create full year
                frac += Integer.parseInt(s.substring(0, 2)) * 100;
                dates.add(frac);
            }
        } else {
            String s = date.replaceAll("[^\\d]", "");
            if (!s.isEmpty()) {
                int base = Integer.parseInt(s);
                dates.add(base);
            }
        }
        return dates;
    }
}
